/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*- */
/*
 * main.c
 * Copyright (C) 2015 Jake R. Capangpangan <camicrisystems@gmail.com>
 *
 * cube-server is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * cube-server is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

using Gee;

public class Main
{
	static CubeServer server;
	static CubeServerRouter router;
	static CubeServerSystem server_system;

	public static int main (string[] args)
	{		
		if (!Thread.supported ()) {
    		stderr.printf ("Cannot run without thread support.\n");
    		return 1;
		}

		if ( initialize_server() )
			server.start_server();
		else
			stdout.printf("Error : Failed to start server!\n");

		return 0;
	}

	public static bool initialize_server()
	{
		uint16 port = 8080;
		string? config_port = "";
		string data_dir = "";
		string projs_dir = "";

		string cube_system_dir_linux = "/usr/share/cube-system";
		string cube_system_dir_win = Path.build_filename(Environment.get_current_dir(),"cube-system");

		if ( SystemInformation.get_operating_system_type() == OperatingSystemType.LINUX )
		{
			data_dir = Path.build_filename( cube_system_dir_linux ,"data");
			projs_dir = Path.build_filename( Environment.get_home_dir() , "cube-system" ,"projects" );
			Environment.set_current_dir( cube_system_dir_linux );
		}
		else if ( SystemInformation.get_operating_system_type() == OperatingSystemType.WINDOWS )
		{
			data_dir = Path.build_filename( cube_system_dir_win ,"data");
			projs_dir = Path.build_filename( cube_system_dir_win ,"projects" );
			Environment.set_current_dir( cube_system_dir_win );
		}

		server_system = new CubeServerSystem();
		router = new CubeServerRouter( server_system );

		if ( !server_system.initialize_cube_system(data_dir,projs_dir) )
		{
			return false;
		}

		config_port = server_system.cube.base_manager.main_configuration_file.get_value("server-port");
		if ( config_port != null )
		{
			port = (uint16)int.parse(config_port);
			if ( port <= 0 )
				port = 8080;
		}

		server = new CubeServer();

		if ( !server.use_port(port) )
		{
			stdout.printf("Error : Unable to use port %d. It is either cube-server is already running or the port is used by another application. Please use another port instead.\n",port);
			return false;
		}
		
		server_system.server_port = port;

		server.started.connect( server_started );
		server.stopped.connect( server_stopped );
		server.error_occurred.connect( server_error_occurred );
		server.request_occurred.connect( request_occurred );

		router.response_occurred.connect ( response_occurred );
		router.stop_server_requested.connect ( ()=> { server.stop_server(); } );

		//Launch browser
		bool result = ProcessManager.run_default_application ( "http://localhost:"+port.to_string() );
		if ( !result )
			stdout.printf("Unable to open web browser. Please do it manually by opening a web browser and navigate to http://localhost:"+port.to_string());		

		return true;
	}

	public static void server_started ()
	{
		stdout.printf("[Server] Started\n");
	}

	public static void server_stopped ()
	{
		stdout.printf("[Server] Stopped\n");
	}

	public static void server_error_occurred ( string message )
	{
		stdout.printf ( "[Server] Error : %s\n", message );
	}

	public static void request_occurred ( string location, TreeMap<string,string> data_map )
	{
		stdout.printf("[Router] Request from : %s\n",data_map["location"]);
		router.route ( location, data_map );
	}

	public static void response_occurred (  uint8[] content_data, int content_length, string type )
	{
		server.write ( content_data, content_length, type );
	}
}