/* -*- Mode: vala; tab-width: 4; intend-tabs-mode: t -*- */
/* cube-vala
 *
 * Copyright (C) Jake R. Capangpangan 2015 <camicrisystems@gmail.com>
 *
cube-vala is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * cube-vala is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

public enum InstallationResultType { SUCCESS = 0, FAILED = 1, NO_TERMINAL = 2, INCOMPLETE = 3, PROHIBITED = 4 }

public class InstallationManager : GLib.Object
{
	Package[] _packages;
	BaseManager _base_mgr;
	SourceManager _source_mgr;
	Project _proj;

	public signal void started();
	public signal void finished();
	public signal void failed(string error_message);

    public InstallationManager (Package[] packages, BaseManager base_manager, SourceManager source_manager, Project project)
	{
        _packages = packages;
		_base_mgr = base_manager;
		_source_mgr = source_manager;
		_proj = project;
    }

	public int start_installation()
	{
		//Ensure that we are in the original computer
		if ( !_proj.is_project_original_computer() )
			return InstallationResultType.PROHIBITED;

		string default_installer = _base_mgr.main_configuration_file.get_value("default-installer");
		if ( default_installer != null )
		{
			if ( default_installer == "apt-get" )
				return _start_apt_installation();
			else
				return _start_dpkg_installation();
		}
		else
			return _start_apt_installation();
	}

	private int _start_apt_installation()
	{
		string install_script = """
#!/bin/bash

cache_dir="%s"
apt_options="%s"
packages="%s"
log_file="%s"
res_file="%s"
lock_file="$cache_dir/lock"

apt-get -o dir::cache::archives="$cache_dir" $apt_options install $packages | tee $log_file
echo Result: $? > $res_file
grep "Need to get 0 B" $log_file
if [[ $? -ne 0 ]] ; then
	echo Incomplete >> $res_file
fi

if [[ -f lock_file ]] ; then
	rm -f lock_file
fi

""";
		string install_log_file = Path.build_filename ( _base_mgr.temporary_directory , "install-log.txt" );
		string install_result_file = Path.build_filename ( _base_mgr.temporary_directory , "install-result.txt" );

		string config_params = _base_mgr.main_configuration_file.get_value("apt-get-parameters");
		config_params = config_params!=null?config_params:"";
		string cache_dir = _proj.packages_directory;
		string packages = "";

		foreach ( Package p in _packages )
		{
			//packages += p.name + "/" + _source_mgr.sources[p.source_index].release + " ";
			packages += p.name + "=" + p.version + " ";
		}

		packages = packages.strip();

		install_script = install_script.printf(cache_dir,config_params,packages,install_log_file, install_result_file );

		return _run(install_script);
	}

	private int _start_dpkg_installation()
	{
		string install_script = "#!/bin/bash\ndpkg %s %s | tee %s\necho Result: $? > %s\n";
		string install_log_file = Path.build_filename ( _base_mgr.temporary_directory , "install-log.txt" );
		string install_result_file = Path.build_filename ( _base_mgr.temporary_directory , "install-result.txt" );

		string config_params = _base_mgr.main_configuration_file.get_value("dpkg-parameters");
		config_params = config_params!=null?config_params:"";
		string packages = "";

		foreach ( Package p in _packages )
		{
			string filename = File.new_for_path( p.filename ).get_basename();
			packages += "\"" + Path.build_filename ( _proj.packages_directory , filename ) + "\" ";
		}
		packages = packages.strip();

		install_script = install_script.printf(config_params,packages,install_log_file, install_result_file);
		return _run(install_script);
	}

	private int _run(string install_script)
	{
		string install_result_file = Path.build_filename ( _base_mgr.temporary_directory , "install-result.txt" );

		var install_script_file = File.new_for_path( Path.build_filename (_base_mgr.temporary_directory, "install.sh") );
		try
		{
			if ( install_script_file.query_exists() )
				install_script_file.delete();
			var stream = new DataOutputStream(install_script_file.create(FileCreateFlags.REPLACE_DESTINATION));
			stream.put_string(install_script);
			stream.close();

			ProcessManager.run_get_status(new string[]{"chmod","+x",Path.build_filename(_base_mgr.temporary_directory,"install.sh")});
			string terminal = find_terminal();
			if ( terminal == null )
			{
				stdout.printf("Error: Unable to find terminal\n");
				return InstallationResultType.NO_TERMINAL;
			}
			//stdout.printf("Terminal to use : %s\n",terminal);

			int res = ProcessManager.run_get_status (new string[]{_base_mgr.root_command_file,terminal,"-e","\""+install_script_file.get_path()+"\""});
			if ( res == 0 )
			{
				var file = File.new_for_path(install_result_file);
				if ( file.query_exists() )
				{
					var instream = new DataInputStream(file.read());
					string line;
					int result = 0;
					bool complete = true;
					while ( (line = instream.read_line(null,null)) != null )
					{
						if ( line.contains("Incomplete") )
							complete = false;
						if ( line.contains("Result") )
							result = int.parse(line.split(":",2)[1].strip());
					}

					if ( !complete )
						return InstallationResultType.INCOMPLETE;
					if ( result != 0 )
						return InstallationResultType.FAILED;
				}
			}
			else
				return InstallationResultType.FAILED;
		}
		catch(Error e)
		{
			return InstallationResultType.FAILED;
		}
		return InstallationResultType.SUCCESS;
	}

	private string? find_terminal()
	{
		string terminals = _base_mgr.main_configuration_file.get_value("terminals");
		if (terminals != null)
		{
			if ( terminals.contains(";") )
			{
				foreach ( string terminal in terminals.split(";",-1) )
				{
					/*
					if ( File.new_for_path(terminal.strip()).query_exists() )
						return terminal;
					*/
					string? path = Which.which(terminal.strip(),_base_mgr);
					if ( path != null )
						return path;
				}
			}
			/*
			if ( File.new_for_path(terminals.strip()).query_exists() )
				return terminals.strip();
			*/
			string? path = Which.which(terminals.strip(),_base_mgr);
			if ( path != null )
				return path;
		}

		return null;
	}
}