/* -*- Mode: vala; tab-width: 4; intend-tabs-mode: t -*- */
/* cube-server
 *
 * Copyright (C) Jake R. Capangpangan 2015 <camicrisystems@gmail.com>
 *
cube-server is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * cube-server is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

using Gee;

public enum DownloadType { PACKAGE, REPOSITORY }
public enum MarkPackageType { TO_INSTALL, TO_DOWNLOAD }
public class DownloadStatus
{				
	public DownloadType download_type;
	public int index;
	public Package[] packages;
	public Source[] sources;
	public int[] progress;
	public DownloadStatusType[] status_type;
	public int overall_progress_success;
	public int overall_progress_failed;
	public string[] transfer_rate;
	public string[] size;
	public string[] error_message;
}

public class CubeServerSystem
{
	CubeSystem _cube;
	PackageQuery _query;
	PackageQuery _curr_query;

	TreeMap<string,bool> _flag_status_map = new TreeMap<string,bool>();
	TreeMap<string,string> _flag_result_map = new TreeMap<string,string>();
	
	ArrayList<string> _marked_to_install_packages = new ArrayList<string>();
	ArrayList<string> _marked_to_download_packages = new ArrayList<string>();

	DownloadStatus? _dl_status = null;

	string _err_message = "";

	int _server_port = 8080;

	//Properties
	public CubeSystem cube { get { return _cube; } set { _cube = value; } }
	public TreeMap<string,bool> flag_status_map { get { return _flag_status_map; } }
	public TreeMap<string,string> flag_result_map { get { return _flag_result_map; } }
	public PackageQuery package_query { get { return _query; } set { _query = value; } }
	public PackageQuery current_query { get { return _curr_query; } set { _curr_query = value; } }
	public DownloadStatus download_status { get { return _dl_status; } }

	public ArrayList<string> marked_to_install_packages { get { return _marked_to_install_packages; } }
	public ArrayList<string> marked_to_download_packages { get { return _marked_to_download_packages; } }

	public int server_port { get { return _server_port; } set { _server_port = value; } } 
	
	public string error_message { get { return _err_message; } }	

	public CubeServerSystem()
	{
		initialize_flags ();
	}

	public void initialize_flags()
	{	
		//All '*ing' are process status which takes time to complete		
		
		_flag_status_map["cube_initialized"] = false;
		
		_flag_status_map["repository_scanning"] = false;
		_flag_status_map["repository_scanned"] = false;
		
		_flag_status_map["repository_downloading"] = false;
		
		_flag_status_map["project_opened"] = false;

		_flag_status_map["project_creating"] = false;
		_flag_status_map["project_created"] = false;
		
		_flag_status_map["project_deleting"] = false;
		_flag_status_map["project_deleted"] = false;
		
		_flag_status_map["project_updating"] = false;

		_flag_status_map["system_updating"] = false;

		_flag_status_map["package_downloading"] = false;

		_flag_status_map["package_installing"] = false;

		_flag_result_map["install_success"] = "";
		_flag_result_map["install_message"] = "";
		
		_flag_result_map["download_success"] = "";
		_flag_result_map["download_message"] = "";
		
		_flag_result_map["update_project_success"] = "";
		_flag_result_map["update_project_message"] = "";
		
		_flag_result_map["update_system_success"] = "";
		_flag_result_map["update_system_message"] = "";		
	}

	public bool initialize_cube_system(string data_dir, string projs_dir)
	{		
		_flag_status_map["cube_initialized"] = false;
	
		_cube = new CubeSystem( data_dir , projs_dir );	
		if ( !_cube.create_and_initialize_system() )
			return false;

		_flag_status_map["cube_initialized"] = true;			

		return true;
	}	

	public bool create_project ( string project )
	{
		_err_message = "";

		_flag_status_map["project_creating"] = true;
		_flag_status_map["project_created"] = false;
		
		if ( !_flag_status_map["cube_initialized"] )
		{
			_err_message = "Failed to initialize cube";
			_flag_status_map["project_creating"] = false;		
			return false;
		}

		if ( File.new_for_path ( Path.build_filename(_cube.base_manager.projects_directory,project) ).query_exists() )
		{
			_err_message = "Unable to create project %s. Project exists".printf(project);
			_flag_status_map["project_creating"] = false;
			return false;
		}
	
		if ( !_cube.create_project(project) )
		{			
			_err_message = "Failed to create project %s.".printf(project);
			_flag_status_map["project_creating"] = false;
			return false;
		}

		_flag_status_map["project_creating"] = true;
		_flag_status_map["project_created"] = true;
		return true;
	}	

	public bool open_project ( string project )
	{		
		_err_message = "";
		
		_flag_status_map["project_opened"] = false;
		
		Project p;
		if ( (p = _cube.find_project(project)) == null )
		{
			_err_message = "Failed to open project %s".printf ( project );			
			return false;
		}
		_cube.open_project( p );
				
		_flag_status_map["project_opened"] = true;		
		
		return true;
	}

	public void thread_scan_repositories(bool status_file_only=false)
	{		
		run_thread ( ()=> {			
			if (!_flag_status_map["repository_scanning"])							
				scan_repositories (status_file_only);			
			return null;
		});
	}

	public bool scan_repositories(bool status_file_only=false)
	{
		_flag_status_map["repository_scanning"] = true;
		_flag_status_map["repository_scanned"] = false;
		
		if ( !_flag_status_map["project_opened"] )
		{
			_err_message = "No opened project";
			_flag_status_map["repository_scanning"] = false;
			return false;
		}

		//TODO : This shouldn't be here :( But what can I do? Im so tired now :/
		//Backup Current Query
		ArrayList<string> bkup_query = new ArrayList<string>();
		int bkup_index = 0;
		if ( _curr_query != null )
		{
			bkup_index = _curr_query.index;
			//Ensure that we are not backuping the main query (all packages), we have that already
			if ( _curr_query.query_type == PackageQueryType.SOME )
			{
				foreach ( Package p in _curr_query.current_package_list )
					bkup_query.add(p.name);
			}
		}

		if ( status_file_only )
			_cube.scan_status_repository();
		else
			_cube.scan_repositories();		
		
		_query = new PackageQuery.package_map ( _cube.repository_manager.available_packages, PackageQueryType.ALL );

		//TODO : This shouldn't be here :( But what can I do? Im so tired now :/
		//Restore Current Query		
		if ( _curr_query != null )
		{			
			ArrayList<Package> restore_query = new ArrayList<Package>();
			//Ensure that we are not restoring the main query (all packages)
			if ( _curr_query.query_type == PackageQueryType.SOME )
			{
				foreach ( string name in bkup_query )
				{
					if ( _cube.repository_manager.available_packages.has_key(name) )						
						restore_query.add(_cube.repository_manager.available_packages[name]);
				}
				_curr_query = new PackageQuery.package_list(restore_query,PackageQueryType.SOME);				
			}
			//It must be the main query, just copy
			else
				_curr_query = _query;

			_curr_query.index = bkup_index;
						
			bkup_query.clear();
			restore_query.clear();
		}

		//Refresh marked packages after scan repositories
		refresh_marked_packages();

		//2 seconds Delay
		Thread.usleep(2000000);
		
		_flag_status_map["repository_scanning"] = false;
		_flag_status_map["repository_scanned"] = true;
		return true;
	}

	public bool thread_download_packages(ArrayList<Package> packages)
	{				
		if (_flag_status_map["package_downloading"])
			return false;

		return run_thread ( ()=> {
			download_packages(packages);	
			return null;
		});		
	}
	
	public void download_packages(ArrayList<Package> packages)
	{		
		_dl_status = new DownloadStatus();
		_dl_status.download_type = DownloadType.PACKAGE;
		_dl_status.packages = packages.to_array();
		_dl_status.status_type = new DownloadStatusType[packages.size];
		_dl_status.progress = new int[packages.size];
		_dl_status.transfer_rate = new string[packages.size];
		_dl_status.size = new string[packages.size];
		_dl_status.error_message = new string[packages.size];

		int overall_progress_success = 0;
		int overall_progress_failed = 0;
		ulong handler = _cube.download_item_status_changed.connect ( (index,type,dl) => {

			_dl_status.index = index;
			_dl_status.status_type[index] = type;
			_dl_status.progress[index] = dl.progress;
			_dl_status.transfer_rate[index] = dl.transfer_rate;
			_dl_status.size[index] = dl.size;

			if ( type == DownloadStatusType.FINISHED )
			{
				overall_progress_success++;
				_dl_status.status_type[index] = DownloadStatusType.FINISHED;
				_dl_status.overall_progress_success = (int)((((float)overall_progress_success)/(float)packages.size)*100.00);
			}
			else if ( type == DownloadStatusType.FAILED )
			{
				overall_progress_failed++;
				_dl_status.status_type[index] = DownloadStatusType.FAILED;
				_dl_status.error_message[index] = dl.error_message;
				_dl_status.overall_progress_failed = (int)((((float)overall_progress_failed)/(float)packages.size)*100.00);
			}
		});		
		
		_flag_status_map["package_downloading"] = true;

		//This also includes mark packages
		_cube.download_packages ( _dl_status.packages );

		if ( _dl_status.overall_progress_failed == 0 )
		{
			_flag_result_map["download_success"] = "true";
			_flag_result_map["download_message"] = "Download Successful";
		}
		else
		{
			_flag_result_map["download_success"] = "false";
			_flag_result_map["download_mesage"] = "failed;Failed to download %d package(s).".printf(overall_progress_failed);			
		}

		_flag_status_map["package_downloading"] = false;

		//2 seconds Delay
		Thread.usleep(2000000);

		//Refresh marked packages after download
		refresh_marked_packages();

		_cube.disconnect(handler);		
	}

	public bool thread_download_repositories(string? ppa = null)
	{				
		return run_thread ( ()=> {					
			if (!_flag_status_map["repository_downloading"])
				download_repositories(ppa);
			return null;
		});
	}

	public void download_repositories(string? ppa = null)
	{
		_dl_status = new DownloadStatus();
		_dl_status.download_type = DownloadType.REPOSITORY;	
		_dl_status.index = -1;

		int overall_progress_success = 0;
		int overall_progress_failed = 0;
		
		ulong handler = _cube.download_item_status_changed.connect ( (index,type,dl) => {			
			_dl_status.index = index;
			_dl_status.status_type[index] = type;
			_dl_status.progress[index] = dl.progress;
			_dl_status.transfer_rate[index] = dl.transfer_rate;
			_dl_status.size[index] = dl.size;

			if ( type == DownloadStatusType.FINISHED )
			{
				overall_progress_success++;
				_dl_status.status_type[index] = DownloadStatusType.FINISHED;
				_dl_status.overall_progress_success = (int)((((float)overall_progress_success)/(float)(cube.source_manager.sources.size-1))*100.00);
			}
			else if ( type == DownloadStatusType.FAILED )
			{
				overall_progress_failed++;
				_dl_status.status_type[index] = DownloadStatusType.FAILED;
				_dl_status.error_message[index] = dl.error_message;
				_dl_status.overall_progress_failed = (int)((((float)overall_progress_failed)/(float)(cube.source_manager.sources.size-1))*100.00);
			}
		});		
		
		_flag_status_map["repository_downloading"] = true;

		ArrayList<Source> sources_to_download = new ArrayList<Source>();
		
		if ( ppa != null )
		{
			foreach ( Source s in _cube.source_manager.get_sources_from_ppa(ppa) )
				sources_to_download.add(s);
		}
		else
		{			
			foreach ( Source s in _cube.source_manager.sources )
			{
				if ( s.filename != "status")
					sources_to_download.add(s);
			}
		}

		_dl_status.sources = sources_to_download.to_array();
		_dl_status.status_type = new DownloadStatusType[_dl_status.sources.length];
		_dl_status.progress = new int[_dl_status.sources.length];
		_dl_status.transfer_rate = new string[_dl_status.sources.length];
		_dl_status.size = new string[_dl_status.sources.length];
		_dl_status.error_message = new string[_dl_status.sources.length];
			
		_cube.download_repositories(_dl_status.sources);
		//_cube.source_manager.scan_sources();

		if ( _dl_status.overall_progress_failed == 0 )
		{
			_flag_result_map["download_success"] = "true";
			_flag_result_map["download_message"] = "Download Successful";
		}
		else
		{
			_flag_result_map["download_success"] = "false";
			_flag_result_map["download_message"] = "Failed to download %d of %d repositories".printf(overall_progress_failed,(cube.source_manager.sources.size-1));
		}

		//2 seconds Delay
		Thread.usleep(2000000);
		
		_flag_status_map["repository_downloading"] = false;

		_cube.disconnect(handler);

		//Scan Repositories after download
		scan_repositories();
	}

	public bool download_stop()
	{
		if ( _flag_status_map["repository_downloading"] || _flag_status_map["package_downloading"] )
		{
			stdout.printf("To Stop!\n");
			_cube.download_manager.stop();
			return true;
		}
		return false;
	}

	public bool thread_install_packages(ArrayList<Package> packages)
	{		
		return run_thread ( ()=> {
			if ( !flag_status_map["package_installing"] )
				install_packages(packages);
			return null;
		});
	}

	public bool install_packages(ArrayList<Package> packages)
	{
		_flag_status_map["package_installing"] = true;		
		bool result = cube.install_packages(packages.to_array());		

		if ( result )
		{
			_flag_result_map["install_success"] = "true";
			_flag_result_map["install_message"] = "Installation Successful";
		}
		else
		{
			_flag_result_map["install_success"] = "false";
			_flag_result_map["install_message"] = "Failed to install packages";			
		}		
		
		_flag_status_map["package_installing"] = false;
		
		return result;
	}

	public bool thread_update_project()
	{		
		return run_thread ( ()=> {
			if ( !flag_status_map["project_updating"] )
				update_project();	
			return null;
		});
	}

	public bool update_project()
	{
		_flag_status_map["project_updating"] = true;
		bool result = cube.update_project();

		if ( result )
		{
			_flag_result_map["update_project_success"] = "true";
			_flag_result_map["update_project_message"] = "Update project successful";
		}
		else
		{
			_flag_result_map["update_project_success"] = "false";
			_flag_result_map["update_project_message"] = "Failed to update project";
		}
		
		_flag_status_map["project_updating"] = false;

		//2 seconds Delay
		Thread.usleep(2000000);
		
		return result;
	}

	public bool thread_update_system()
	{		
		return run_thread ( ()=> {
			if ( !flag_status_map["system_updating"] )
				update_system();
			return null;
		});
	}

	public bool update_system()
	{		
		_flag_status_map["system_updating"] = true;
		bool result = cube.update_system();

		if ( result )
		{
			_flag_result_map["update_system_success"] = "true";
			_flag_result_map["update_system_message"] = "Update system successful";
		}
		else
		{
			_flag_result_map["update_system_success"] = "false";
			_flag_result_map["update_system_message"] = "Failed to update system";
		}
		
		_flag_status_map["system_updating"] = false;

		//2 seconds Delay
		Thread.usleep(2000000);
		
		return result;
	}

	public void mark_package(MarkPackageType type, string name)
	{
		if ( type == MarkPackageType.TO_INSTALL )
		{
			if ( !_marked_to_install_packages.contains(name) && cube.repository_manager.available_packages.has_key(name) )
				_marked_to_install_packages.add(name);
		}
		else if ( type == MarkPackageType.TO_DOWNLOAD )
		{
			if ( !_marked_to_download_packages.contains(name) && cube.repository_manager.available_packages.has_key(name) )
				_marked_to_download_packages.add(name);
		}		
	}

	public void unmark_package(MarkPackageType type, string name)
	{
		if ( type == MarkPackageType.TO_INSTALL )
		{
			if ( _marked_to_install_packages.contains(name) )
				_marked_to_install_packages.remove(name);
		}
		else if ( type == MarkPackageType.TO_DOWNLOAD )
		{
			if ( _marked_to_download_packages.contains(name) )
				_marked_to_download_packages.remove(name);
		}		
	}

	public void clear_marked_packages(MarkPackageType type)
	{
		if ( type == MarkPackageType.TO_INSTALL )
			_marked_to_install_packages.clear();
		else if ( type == MarkPackageType.TO_DOWNLOAD )
			_marked_to_download_packages.clear();
	}

	public void refresh_marked_packages()
	{
		string[] marked = _marked_to_install_packages.to_array();
		
		for ( int i = 0; i < marked.length; i++ )
		{
			if ( _cube.repository_manager.available_packages[marked[i]].status == PackageStatusType.INSTALLED )
				_marked_to_install_packages.remove(marked[i]);			
		}

		marked = _marked_to_download_packages.to_array();

		for ( int i = 0; i < marked.length; i++ )
		{
			if ( _cube.repository_manager.available_packages[marked[i]].status == PackageStatusType.DOWNLOADED )
				_marked_to_download_packages.remove(marked[i]);
		}
	}

	public bool run_thread(ThreadFunc<void*> run)
	{
		try
		{
			unowned Thread<void*> thread = Thread.create<void*>(run,true);
		}
		catch ( Error e )
		{
			stdout.printf("Error : %s\n",e.message);
			return false;
		}
		return true;
	}

	public string view_headers ( string title )
	{
		string content =
		"""
<!--Generated by Camicri Cube Server 2.0-->
<!DOCTYPE html>
<html lang="en">
	<head>
		<meta charset="utf-8">
		<meta http-equiv="X-UA-Compatible" content="IE=edge">
		<meta name="viewport" content="width=device-width, initial-scale=1">
		<title>%s</title>
		<link href="/assets/css/bootstrap.min.css" rel="stylesheet">		
	</head>
<body>

<!--Starts Here-->
		""".printf(title);
		return content;
	}

	public string view_footers ( string extra = "" )
	{
		string content =
		"""
<!--Ends Here-->
</body>
		<script type='text/javascript' src="/assets/js/jquery.min.js"></script>
		<script type='text/javascript' src="/assets/js/bootstrap.min.js"></script>
		%s		
</html>
		""";
		return content.printf(extra);
	}
}